from mconfig import MCONFIG
from rich import print

import motion

# Test that for simple queries, the results make some sense


def test_scrape() -> None:
    connection = motion.test(
        MCONFIG,
        wait_for_triggers=["ScrapeBonAppetit"],
        motion_logging_level="INFO",  # Can be "INFO" or "DEBUG" for more verbose logging
        session_id="EXAMPLE_SESSION_ID",  # Can comment this out to generate a new session ID
    )
    print(f"Session ID: {connection.session_id}")

    # Must specify kw for every arg in .set and .get
    ingredients = "pasta;tomatoes;garlic;cheese"
    new_id = connection.set(
        relation="Query",
        identifier="",
        key_values={"ingredients": ingredients},
    )
    recipe_ids_and_scores = connection.get(
        relation="Query",
        identifier=new_id,
        keys=["recipe_id", "recipe_score"],
        include_derived=True,
        as_df=True,
    )
    recipe_ids_and_titles = connection.mget(
        relation="Recipe",
        identifiers=list(recipe_ids_and_scores["recipe_id"].values),
        keys=["title", "recipe_url"],
        as_df=True,
    ).rename(columns={"identifier": "recipe_id"})
    result = recipe_ids_and_scores.merge(recipe_ids_and_titles, on="recipe_id")

    print(f"Ingredient list: {ingredients}")
    print(f"Response: {result}")


test_scrape()



import os
import shutil

import click

import motion


@click.group()
def motioncli() -> None:
    pass


@motioncli.command("example")
@click.option(
    "--name",
    prompt="Example application name",
    help="One of the example applications to create. Can be: 'cooking'.",
)
@click.option(
    "--author",
    prompt="Your name",
    help="Author name.",
)
def example(name: str, author: str) -> None:
    """Creates a new application from an example."""
    motion.create_example_app(name, author)
    click.echo("Created a project successfully.")


@motioncli.command("create")
@click.option(
    "--name",
    prompt="Your application name",
    help="The name of your application.",
)
@click.option(
    "--author",
    prompt="Your name",
    help="Author name.",
)
def create(name: str, author: str) -> None:
    """Creates a new application."""
    motion.create_app(name, author)

    click.echo("Created a project successfully.")


@motioncli.command("serve")
@click.argument("host", required=False, default="0.0.0.0")
@click.argument("port", required=False, default=5000)
@click.option(
    "logging_level",
    "--logging-level",
    "-l",
    default="WARNING",
    help="Logging level for motion. Can be DEBUG, INFO, WARNING, ERROR, CRITICAL.",
)
def serve(host: str, port: int, logging_level: str) -> None:
    """Serves a Motion application."""

    # Check that the project is created
    if not os.path.exists("mconfig.py"):
        click.echo("Project is not created. Run `motion create` first.")
        return

    # Create object from mconfig.py
    config_code = open("mconfig.py").read() + "\nMCONFIG"

    import sys

    sys.path.append(os.getcwd())

    exec(config_code)
    mconfig = locals()["MCONFIG"]
    click.echo(f"Serving application {mconfig['application']['name']}...")

    # Serve the application
    motion.serve(mconfig, host=host, port=port, motion_logging_level=logging_level)


@motioncli.command("clear")
@click.argument("name", required=True)
def clear(name: str) -> None:
    """Removes the data store for the given Motion application."""
    # Remove directory at name
    MOTION_HOME = os.environ.get("MOTION_HOME", os.path.expanduser("~/.cache/motion"))
    dirname = os.path.join(MOTION_HOME, "datastores", name)
    if not os.path.exists(dirname):
        click.echo(f"Application {name} does not exist.")
        return

    shutil.rmtree(dirname)


@motioncli.command("token", help="Generate a new API token")
def token() -> None:
    """Generate a new API token."""
    token = motion.create_token()
    click.echo(token)



import os
import threading
import typing

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from croniter import croniter

from motion import Schema, Trigger
from motion.cursor import Cursor
from motion.task import CheckpointThread, CronThread
from motion.utils import TriggerFn, logger


class Store:
    def __init__(
        self,
        name: str,
        session_id: str,
        datastore_prefix: str = "datastores",
        checkpoint: str = "0 * * * *",
        disable_cron_triggers: bool = False,
    ) -> None:
        self.name = name
        self.session_id: str = session_id

        self.datastore_prefix = datastore_prefix
        self.checkpoint_interval = checkpoint
        self.disable_cron_triggers = disable_cron_triggers

        # Set listening to false
        self._listening = False

        # self.con = duckdb.connect(":memory:")
        if not os.path.exists(os.path.join(datastore_prefix, self.name)):
            os.makedirs(os.path.join(datastore_prefix, self.name))

        self.db_write_lock = threading.Lock()

        # Try loading from checkpoint
        (
            self.relations,
            self.table_columns,
            self.log_table,
        ) = self.loadFromCheckpoint_pa()
        if self.log_table is None:
            self.addLogTable_pa()

        # Set up triggers
        self.triggers: typing.Dict[str, list[TriggerFn]] = {}
        self.cron_triggers: typing.Dict[str, list[TriggerFn]] = {}
        self.cron_threads: typing.Dict[str, CronThread] = {}
        self.trigger_names: typing.Dict[str, list[str]] = {}
        self.trigger_fns: typing.Dict[str, Trigger] = {}

    def __del__(self) -> None:
        self.stop(wait=False)

    @property
    def listening(self) -> bool:
        return self._listening

    def cursor(
        self, bypass_listening: bool = False, wait_for_results: bool = False
    ) -> Cursor:
        """Generates a new cursor for the database, with triggers and all.

        Returns:
            Cursor: The cursor.
        """
        if not self.listening and not bypass_listening:
            raise Exception(
                "Store has not started. Call store.start() before using the cursor."
            )
        return Cursor(
            name=self.name,
            relations=self.relations,
            log_table=self.log_table,
            table_columns=self.table_columns,
            triggers=self.triggers,
            write_lock=self.db_write_lock,
            session_id=self.session_id,
            wait_for_results=wait_for_results,
        )

    def checkpoint_pa(self) -> None:
        """Checkpoint store object."""
        # Save relations
        base_path = os.path.join(self.datastore_prefix, self.name)
        for relation in self.relations:
            os.makedirs(os.path.join(base_path, relation), exist_ok=True)
            ds.write_dataset(
                self.relations[relation],
                base_dir=os.path.join(base_path, relation),
                format="parquet",
                partitioning=ds.partitioning(
                    pa.schema([("session_id", pa.string())]), flavor="hive"
                ),
                existing_data_behavior="delete_matching",
                schema=self.relations[relation].schema,
            )

        # Save logs
        pq.write_table(self.log_table, os.path.join(base_path, "logs.parquet"))

        # TODO: checkpoint trigger objects

    def loadFromCheckpoint_pa(self) -> tuple[dict, dict, pa.Table]:
        """Load store object from checkpoint."""
        try:
            relations = {}
            table_columns = {}

            base_path = os.path.join(self.datastore_prefix, self.name)

            # Iterate through all folders in base_path
            for relation in os.listdir(base_path):
                if relation == "logs.parquet":
                    continue

                dataset = ds.dataset(
                    os.path.join(base_path, relation),
                    format="parquet",
                    partitioning="hive",
                )

                # Load session_id partition
                try:
                    table = dataset.to_table(
                        filter=pc.equal(ds.field("session_id"), self.session_id)
                    )
                except pa.ArrowInvalid as e:
                    # If no session_id partition exists, move on
                    continue

                relations[relation] = table

                # Load table columns
                table_columns[relation] = table.schema.names
                table_columns[relation].remove("identifier")
                table_columns[relation].remove("derived_id")

                logger.info(
                    f"Loaded relation {relation} from checkpoint with {table.num_rows} existing rows in session."
                )

            # Load logs
            log_table = pq.read_table(os.path.join(base_path, "logs.parquet"))

            return relations, table_columns, log_table

        except Exception as e:
            logger.warning(
                f"Could not load database {self.name} from checkpoint. Error: {e}"
            )
            return {}, {}, None

        # TODO: load trigger objects

    def addLogTable_pa(self) -> None:
        """Creates a table to store trigger logs."""

        schema = pa.schema(
            [
                pa.field(
                    "executed_time",
                    pa.timestamp("us"),
                    nullable=False,
                ),
                pa.field("session_id", pa.string(), nullable=False),
                pa.field("trigger_name", pa.string(), nullable=False),
                pa.field("trigger_version", pa.int64(), nullable=False),
                pa.field("trigger_action", pa.string(), nullable=False),
                pa.field("relation", pa.string(), nullable=False),
                pa.field("identifier", pa.string(), nullable=False),
                pa.field("trigger_key", pa.string(), nullable=False),
            ]
        )
        # Create table with schema
        self.log_table = schema.empty_table()

    def addrelation_pa(self, name: str, schema: Schema) -> None:
        """_Add a relation to the store.

        Args:
            name (str): The name of the relation.
            schema (motion.Schema): The schema of the relation.
        """
        pa_schema = schema.formatPaSchema(name)

        if name in self.relations:
            pa_schema_names = pa_schema.names
            pa_schema_types = pa_schema.types

            for old_name, old_type in zip(
                self.relations[name].schema.names,
                self.relations[name].schema.types,
            ):
                if old_name == "session_id":
                    continue
                name_idx = pa_schema_names.index(old_name)
                if not old_type.equals(pa_schema_types[name_idx]):
                    logger.error(
                        f"relation {name} already exists with a different schema. Please clear the data store with `motion clear {self.name}` and try again."
                    )

        else:
            logger.info(f"Adding relation {name} with schema {pa_schema}")
            self.relations[name] = pa_schema.empty_table()

            self.table_columns[name] = self.relations[name].schema.names
            self.table_columns[name].remove("identifier")
            self.table_columns[name].remove("derived_id")

    def addTrigger(
        self,
        name: str,
        trigger: typing.Type[Trigger],
        params: typing.Dict[str, typing.Any] = {},
    ) -> None:
        """Adds a trigger to the store.

        Args:
            name (str): Trigger name.
            trigger (Trigger): Trigger class to execute when trigger is fired. Must implement the Trigger interface.
            params (typing.Dict[str, typing.Any], optional): Parameters to pass

        Raises:
            ValueError: If there is already a trigger with the given name.
        """
        if name in self.trigger_names:
            logger.warning(f"Trigger {name} already exists. Doing nothing.")
            return

        # Check that the class implements the Trigger interface
        if not issubclass(trigger, Trigger):
            raise ValueError(f"Trigger class must implement the Trigger interface.")

        # Retrieve keys
        keys = trigger.getRouteKeys()

        # Check that keys are valid
        all_possible_keys = [
            f"{ns}.{key}" for ns in self.table_columns for key in self.table_columns[ns]
        ]
        cron_key_exists = False
        for full_key in keys:
            _, key = full_key.split(".")
            if full_key not in all_possible_keys and not croniter.is_valid(key):
                raise ValueError(
                    f"Trigger {name} has invalid key {full_key}. Valid keys are {all_possible_keys} or a cron expression. If your schemas have changed, you may need to clear your application by running `motion clear {self.name}`."
                )

            if croniter.is_valid(key):
                if cron_key_exists:
                    raise ValueError(
                        f"Trigger {name} has more than one cron key. Only one cron key is allowed per trigger."
                    )

                cron_key_exists = True

        # Add the trigger to the store
        self.trigger_names[name] = keys

        version = pc.max(
            pc.filter(
                self.log_table["trigger_version"],
                pc.equal(self.log_table["trigger_name"], name),
            )
        ).as_py()

        version = version if version is not None else 0

        trigger_exec = trigger(
            self.cursor(bypass_listening=True), name, version, params
        )
        self.trigger_fns[name] = trigger_exec

        for full_key in keys:
            _, key = full_key.split(".")
            if croniter.is_valid(key):
                self.cron_triggers[full_key] = self.cron_triggers.get(full_key, []) + [
                    TriggerFn(name, trigger_exec)
                ]

            else:
                self.triggers[full_key] = self.triggers.get(full_key, []) + [
                    TriggerFn(name, trigger_exec)
                ]

    def deleteTrigger(self, name: str) -> None:
        """Delete a trigger from the store.

        Args:
            name (str): The name of the trigger.
        """
        if name not in self.trigger_names:
            raise ValueError(f"Trigger {name} does not exist.")

        # Remove the trigger from the store
        keys = self.trigger_names[name]
        fn = self.trigger_fns[name]
        for key in keys:
            if croniter.is_valid(key):
                self.cron_triggers[key].remove(TriggerFn(name, fn))
                self.cron_threads[name].stop()
                self.cron_threads[name].join()

            else:
                self.triggers[key].remove(TriggerFn(name, fn))

        del self.trigger_names[name]
        del self.trigger_fns[name]

    def getTriggersForKey(self, relation: str, key: str) -> list[str]:
        """Get the list of triggers for a given key.

        Args:
            relation (str): The relation to get the triggers for.
            key (str): The key to get the triggers for.

        Returns:
            typing.List[str]: The list of triggers for the given key.
        """
        names_and_fns = self.triggers.get(f"{relation}.{key}", [])
        return [t[0] for t in names_and_fns]

    def getTriggersForAllKeys(self) -> typing.Dict[str, list[str]]:
        """Get the list of triggers for all keys.

        Returns:
            typing.Dict[str, typing.List[str]]: The list of triggers for all keys.
        """
        return {
            k: self.getTriggersForKey(k.split(".")[0], k.split(".")[1])
            for k in self.triggers.keys()
        }

    def start(self) -> None:
        """Start the store."""
        # Start cron triggers
        self._listening = True
        self.cron_threads = {}

        if not self.disable_cron_triggers:
            for full_key, triggers in self.cron_triggers.items():
                _, cron_expression = full_key.split(".")
                for trigger_fn in triggers:
                    e = threading.Event()
                    t = CronThread(
                        cron_expression,
                        self.cursor(wait_for_results=True),
                        trigger_fn,
                        self.checkpoint_pa,
                        e,
                        self.session_id,
                    )
                    self.cron_threads[trigger_fn.name] = t
                    t.start()

        # Start a thread to checkpoint the store every 5 minutes
        self.checkpoint_thread = CheckpointThread(
            self.name, self.checkpoint_pa, self.checkpoint_interval
        )
        self.checkpoint_thread.start()

    def waitForTrigger(self, trigger_name: str) -> None:
        """Wait for a cron-scheduled trigger to fire.

        Args:
            trigger_name (str): The name of the trigger to wait for.
        """
        if self.disable_cron_triggers:
            raise ValueError(
                f"Cannot wait for trigger {trigger_name} because cron triggers are disabled."
            )

        if trigger_name not in self.cron_threads.keys():
            raise FileNotFoundError(
                f"Trigger {trigger_name} does not exist as a cron-scheduled thread. Valid cron-scheduled triggers are {list(self.cron_threads.keys())}."
            )

        logger.info(f"Waiting for trigger {trigger_name} to fire...")
        self.cron_threads[trigger_name].first_run_event.wait()

    def stop(self, wait: bool = False) -> None:
        """Stop the store."""
        # Stop cron triggers

        if self._listening:
            for _, t in self.cron_threads.items():
                t.stop()
                if wait:
                    t.join()

            self.checkpoint_thread.stop()
            if wait:
                self.checkpoint_thread.join()

        self._listening = False

        logger.info("Stopped store.")



import inspect
import sys
import threading
import typing
from abc import ABC, abstractmethod
from queue import SimpleQueue

from motion.cursor import Cursor
from motion.utils import TriggerElement, logger


class CustomDict(dict):
    def __init__(
        self,
        trigger_name: str,
        dict_type: str,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        self.trigger_name = trigger_name
        self.dict_type = dict_type
        super().__init__(*args, **kwargs)

    def __getitem__(self, key: str) -> object:
        try:
            return super().__getitem__(key)
        except KeyError:
            raise KeyError(
                f"Key `{key}` not found in {self.dict_type} for trigger {self.trigger_name}."
            )


class Trigger(ABC):
    """A trigger is a class that defines the logic for a particular type of inference. Triggers are instantiated once per project and are responsible for maintaining their own state. Triggers are responsible for defining their setup, routes that they respond to, and the logic for infer and fit. See the [trigger life cycle](/concepts/trigger) for more information."""

    def __init__(
        self,
        cursor: Cursor,
        name: str,
        version: int,
        params: dict = {},
        routes_only: bool = False,
    ):
        self.name = name

        # Validate number of arguments in each trigger and set up routes
        route_list = self.routes()
        if not isinstance(route_list, list):
            raise TypeError(
                f"routes() of trigger {name} should return a list of motion.Route objects."
            )

        seen_keys = set()
        for r in route_list:
            if f"{r.relation}.{r.key}" in seen_keys:
                raise ValueError(
                    f"Duplicate route {r.relation}.{r.key} in trigger {name}."
                )

            r.validateTrigger(self)
            seen_keys.add(f"{r.relation}.{r.key}")

        self.route_map = {}
        for r in self.routes():
            if r.relation != "":
                self.route_map[f"{r.relation}.{r.key}"] = r
            else:
                self.route_map[f"_cron.{r.key}"] = r

        if routes_only:
            return

        # Set up params dictionary
        self._params = CustomDict(self.name, "params", params)

        # Set up initial state
        if len(inspect.signature(self.setUp).parameters) != 1:
            raise ValueError(f"setUp() of trigger {name} should have 1 argument")

        self._state = CustomDict(self.name, "state", {})
        self._version = version
        self._last_fit_id = -sys.maxsize - 1

        initial_state = self.setUp(cursor)
        if not isinstance(initial_state, dict):
            raise TypeError(f"setUp() of trigger {self.name} should return a dict.")
        self.update(initial_state)

        # Set up fit queue
        self._fit_queue = SimpleQueue()  # type: SimpleQueue
        self._fit_thread = threading.Thread(
            target=self.processFitQueue,
            daemon=True,
            name=f"{name}_fit_thread",
        )
        self._fit_thread.start()

    @classmethod
    def getRouteKeys(cls) -> list:
        obj: Trigger = cls(None, "", 0, routes_only=True)  # type: ignore

        return list(obj.route_map.keys())

    @abstractmethod
    def routes(self) -> list:
        """Specifies mappings from trigger keys to lifecycle functions (i.e., infer, fit methods).
        Returns:
            list: List of routes that this trigger responds to. Each route is a motion.Route object.
        """
        pass

    @abstractmethod
    def setUp(self, cursor: Cursor) -> dict:
        """Sets up the initial state of the trigger. Called only when the application is started.

        Args:
            cursor (Cursor): Cursor object to access the Motion data store.

        Raises:
            NotImplementedError: Error if this method is not implemented.

        Returns:
            dict: Initial state of the trigger.
        """
        raise NotImplementedError(f"Please implement setUp() for trigger {self.name}.")

    @property
    def params(self) -> dict:
        return self._params

    @property
    def state(self) -> dict:
        return self._state

    @property
    def version(self) -> int:
        return self._version

    @property
    def last_fit_id(self) -> int:
        return self._last_fit_id

    def update(self, new_state: dict) -> None:
        if new_state:
            self._state.update(new_state)
            self._version += 1

    def processFitQueue(self) -> None:
        while True:
            (
                cursor,
                trigger_name,
                trigger_context,
                infer_context,
                fit_event,
            ) = self._fit_queue.get()

            new_state = self.route_map[
                f"{trigger_context.relation}.{trigger_context.key}"
            ].fit(cursor, trigger_context, infer_context)

            if not isinstance(new_state, dict):
                fit_event.set()
                raise TypeError(
                    f"fit() of trigger {self.name} should return a dict of state updates."
                )

            old_version = self.version
            self.update(new_state)

            logger.info(
                f"Finished running trigger {trigger_name} for identifier {trigger_context.identifier} and key {trigger_context.key}."
            )

            cursor.logTriggerExecution(
                trigger_name, old_version, "fit", trigger_context
            )

            fit_event.set()

    def fitWrapper(
        self,
        cursor: Cursor,
        trigger_name: str,
        trigger_context: TriggerElement,
        infer_context: typing.Any,
    ) -> threading.Event:
        fit_event = threading.Event()
        self._fit_queue.put(
            (cursor, trigger_name, trigger_context, infer_context, fit_event)
        )

        return fit_event



from motion.trigger import Trigger
from motion.schema import Schema
from motion.entry import (
    init,
    serve,
    connect,
    test,
    create_token,
    create_app,
    create_example_app,
)
from motion.routing import Route
from motion.utils import update_params
from motion.client import ClientConnection
from motion.cursor import Cursor
from motion.utils import TriggerElement

__all__ = [
    "Schema",
    "Trigger",
    "init",
    "serve",
    "connect",
    "test",
    "Route",
    "update_params",
    "ClientConnection",
    "create_token",
    "create_app",
    "create_example_app",
    "Cursor",
    "TriggerElement",
]



import pandas as pd
import streamlit as st
from mconfig import MCONFIG

import motion


@st.cache_resource
def setup_database() -> motion.ClientConnection:
    connection = motion.test(
        MCONFIG,
        wait_for_triggers=["ScrapeBonAppetit"],
        # motion_logging_level="WARNING",  # Can be "INFO" or "DEBUG" for more verbose logging
        session_id="EXAMPLE_SESSION_ID",
    )
    return connection


connection = setup_database()


@st.cache_data(show_spinner="Fetching results...")
def run_query(ingredients: str) -> pd.DataFrame:
    # Retrieve the results and get the lowest cosine similarity
    # (i.e., best match) for each img_id
    new_id = connection.set(
        relation="Query",
        identifier="",
        key_values={"ingredients": ingredients},
    )
    recipe_ids_and_scores = connection.get(
        relation="Query",
        identifier=new_id,
        keys=["recipe_id", "recipe_score"],
        include_derived=True,
        as_df=True,
    )
    recipe_ids_and_titles = connection.mget(
        relation="Recipe",
        identifiers=list(recipe_ids_and_scores["recipe_id"].values),
        keys=["title", "recipe_url", "image_url"],
        as_df=True,
    ).rename(columns={"identifier": "recipe_id"})
    results = recipe_ids_and_scores.merge(recipe_ids_and_titles, on="recipe_id")

    return results


ingredients = st.text_input("Type a list of ingredients you have in your fridge")

if ingredients:
    image_results = run_query(ingredients)

    #  with st.spinner("Fetching results..."):
    st_cols = st.columns(3)
    col_idx = 0

    for _, row in image_results.iterrows():
        st_cols[col_idx].image(row["image_url"])
        st_cols[col_idx].markdown(f'{row["title"]} | [Link]({row["recipe_url"]})')
        col_idx = (col_idx + 1) % 3



import os
import typing
from io import BytesIO
from urllib.parse import parse_qs

import pandas as pd
from fastapi import Depends, FastAPI, File, HTTPException, Request, Response, UploadFile
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from pydantic import Json

from motion.api.models import *
from motion.store import Store


def df_to_json_response(df: pd.DataFrame) -> Response:
    return Response(
        df.to_parquet(engine="pyarrow", index=False),
        media_type="application/octet-stream",
    )


def create_fastapi_app(store: Store, testing: bool = False) -> FastAPI:
    # Security
    MOTION_API_TOKEN = os.environ.get("MOTION_API_TOKEN")
    if not MOTION_API_TOKEN:
        raise ValueError(
            "MOTION_API_TOKEN environment variable is not set. Please run `motion token` to generate a token, and export MOTION_API_TOKEN={result}."
        )

    scheme = HTTPBearer()

    def check_auth(
        credentials: HTTPAuthorizationCredentials = Depends(scheme),
    ) -> bool:
        if credentials.credentials != MOTION_API_TOKEN:
            raise HTTPException(
                status_code=401, detail="Invalid authentication credentials"
            )

        return True

    app = FastAPI(dependencies=[Depends(check_auth)])

    json_app = FastAPI(
        title="Motion JSON API",
        description="An API for calling basic Motion functions on JSON data.",
        dependencies=[Depends(check_auth)]
        # version="1.0.0",
        # servers=[{"url": "https://your-app-url.com"}],
    )
    app.mount("/json", json_app)

    @app.on_event("startup")
    async def startup() -> None:
        app.state.testing = testing
        app.state.store = store

    @app.get("/get/")
    async def get(args: GetRequest) -> Response:
        cur = app.state.store.cursor()
        res = cur.get(**args.__dict__)

        # Check if the result is a pandas df
        if isinstance(res, dict):
            res = pd.DataFrame(res, index=[0])

        elif not isinstance(res, pd.DataFrame):
            res = pd.DataFrame(res)

        return df_to_json_response(res)

    @app.get("/mget/")
    async def mget(args: MgetRequest) -> Response:
        cur = app.state.store.cursor()
        res = cur.mget(
            **args.__dict__,
        )
        # Check if the result is a pandas df
        if isinstance(res, dict):
            res = pd.DataFrame(res, index=[0])

        elif not isinstance(res, pd.DataFrame):
            res = pd.DataFrame(res)

        return df_to_json_response(res)

    @app.post("/set_python/")
    async def set_python(
        args: Json[PartialSetRequest], key_values: UploadFile = File(...)
    ) -> typing.Any:
        args = args.dict()  # type: ignore
        content = await key_values.read()
        with BytesIO(content) as f:
            if key_values.content_type == "application/octet-stream":
                df = pd.read_parquet(f, engine="pyarrow")
                args["key_values"] = df.to_dict(orient="records")[0]  # type: ignore
            else:
                raise HTTPException(status_code=400, detail="Invalid file type")

        cur = app.state.store.cursor()
        return cur.set(
            relation=args["relation"],  # type: ignore
            identifier=args["identifier"],  # type: ignore
            key_values=args["key_values"],  # type: ignore
        )

    @app.get("/sql/")
    async def sql(args: SqlRequest) -> Response:
        cur = app.state.store.cursor()
        res = cur.sql(args.query, args.as_df)
        if isinstance(res, pd.DataFrame):
            return df_to_json_response(res)
        else:
            return Response(res, media_type="application/json")

    @app.post("/duplicate/")
    async def duplicate(data: DuplicateRequest) -> typing.Any:
        cur = app.state.store.cursor()
        return cur.duplicate(relation=data.relation, identifier=data.identifier)

    @app.post("/wait_for_trigger/")
    async def wait_for_trigger(request: Request) -> str:
        data = await request.body()
        trigger = parse_qs(data.decode("unicode_escape"))["trigger"][0]

        app.state.store.waitForTrigger(trigger)
        return trigger

    @app.get("/ping/")
    async def root() -> dict:
        return {"message": "Hello World"}

    @app.get("/session_id/")
    async def session_id() -> typing.Any:
        return app.state.store.session_id

    @app.post("/checkpoint/")
    async def checkpoint() -> typing.Any:
        return app.state.store.checkpoint_pa()

    @app.on_event("shutdown")
    async def shutdown() -> None:
        if not app.state.testing:
            app.state.store.stop()

    # JSON API

    @json_app.post("/set/")
    async def json_set(args: SetRequest) -> typing.Any:
        cur = app.state.store.cursor()
        try:
            identifier = cur.set(
                relation=args.relation,
                identifier=args.identifier,
                key_values=args.key_values,
            )
            return identifier
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @json_app.get("/get/")
    async def json_get(args: GetRequest) -> typing.Any:
        cur = app.state.store.cursor()

        try:
            args = args.dict()  # type: ignore
            args["as_df"] = False  # type: ignore
            res = cur.get(**args)
            return res
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @json_app.get("/mget/")
    async def json_mget(args: MgetRequest) -> typing.Any:
        cur = app.state.store.cursor()
        try:
            args = args.dict()  # type: ignore
            args["as_df"] = False  # type: ignore
            res = cur.mget(**args)
            return res
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @json_app.get("/sql/")
    async def json_sql(args: SqlRequest) -> typing.Any:
        cur = app.state.store.cursor()
        try:
            res = cur.sql(args.query, as_df=True)
            return res.to_dict(orient="records")
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @json_app.get("/session_id/")
    async def json_session_id() -> typing.Any:
        return app.state.store.session_id

    @json_app.post("/wait_for_trigger/")
    async def json_wait_for_trigger(args: WaitRequest) -> typing.Any:
        cur = app.state.store.cursor()
        try:
            app.state.store.waitForTrigger(args.trigger)
            return args.trigger
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    return app



import os
import typing
from collections import namedtuple
from typing import Dict, List, Tuple

import cohere
import faiss
import numpy as np

import motion

IngredientsList = namedtuple("IngredientsList", ["identifier", "ingredients"])


class SearchRecipe(motion.Trigger):
    def routes(self) -> List[motion.Route]:
        return [
            motion.Route(
                relation="Recipe",
                key="ingredients",
                infer=None,
                fit=self.addRecipeToIndex,
            ),
            motion.Route(
                relation="Query",
                key="ingredients",
                infer=self.findNearestRecipes,
                fit=None,
            ),
        ]

    def setUp(self, cursor: motion.Cursor) -> Dict:
        # Set up the recipe index and cohere client
        co = cohere.Client(os.environ["COHERE_API_KEY"])
        recipe_index = faiss.IndexFlatIP(4096)
        recipe_index_to_id: Dict[int, str] = {}

        ingredients_df = cursor.sql("SELECT identifier, ingredients FROM Recipe")

        if len(ingredients_df) > 0:
            ingredients_list = ingredients_df["ingredients"].tolist()
            recipe_ids = ingredients_df["identifier"].tolist()

            recipe_index, recipe_index_to_id = self._populateRecipeIndex(
                co,
                ingredients_list,
                recipe_ids,
                recipe_index,
                recipe_index_to_id,
            )

        return {
            "cohere": co,
            "recipe_stream": [],
            "recipe_index": recipe_index,
            "recipe_index_to_id": recipe_index_to_id,
        }

    def _populateRecipeIndex(
        self,
        co: cohere.Client,
        ingredients_list: List[str],
        recipe_ids: List[str],
        recipe_index: faiss.IndexFlatIP,
        recipe_index_to_id: Dict[int, str],
    ) -> Tuple[faiss.IndexFlatIP, Dict[int, str]]:
        response = co.embed(
            texts=ingredients_list,
        ).embeddings
        embedding_list = np.array(response)
        # Normalize the embeddings
        embedding_list = embedding_list / np.linalg.norm(
            embedding_list, axis=1, keepdims=True
        )

        # Add the embeddings to the index
        recipe_index.add(embedding_list)
        for rid in recipe_ids:
            recipe_index_to_id[len(recipe_index_to_id)] = rid

        return recipe_index, recipe_index_to_id

    def addRecipeToIndex(
        self,
        cursor: motion.Cursor,
        trigger_context: motion.TriggerElement,
        infer_context: typing.Any,
    ) -> Dict:
        # Keep a stream of the last 20 recipes,
        # adding them to the index every 20 iterations.
        # This is to speed up the process of adding recipes

        recipe_stream = self.state["recipe_stream"]
        recipe_stream.append(
            IngredientsList(trigger_context.identifier, trigger_context.value)
        )
        new_state = {"recipe_stream": recipe_stream}

        # Every 20 recipes, add them to the index
        if len(recipe_stream) % 20 == 0:
            # Get the embeddings
            ingredients_list = [r.ingredients for r in recipe_stream]
            recipe_ids = [r.identifier for r in recipe_stream]
            recipe_index, recipe_index_to_id = self._populateRecipeIndex(
                self.state["cohere"],
                ingredients_list,
                recipe_ids,
                self.state["recipe_index"],
                self.state["recipe_index_to_id"],
            )

            new_state.update(
                {
                    "recipe_stream": [],
                    "recipe_index": recipe_index,
                    "recipe_index_to_id": recipe_index_to_id,
                }
            )

        return new_state

    def _searchIndex(self, features: np.ndarray) -> Tuple[np.ndarray, List[str]]:
        scores, indices = self.state["recipe_index"].search(
            features / np.linalg.norm(features, axis=1),
            self.params["num_recipe_results"],
        )
        recipe_ids = [self.state["recipe_index_to_id"][index] for index in indices[0]]
        return scores[0], recipe_ids

    def findNearestRecipes(
        self, cursor: motion.Cursor, trigger_context: motion.TriggerElement
    ) -> None:
        # Find the nearest recipes for the query
        ingredients = trigger_context.value
        response = self.state["cohere"].embed(texts=[ingredients])
        embedding = np.array(response.embeddings[0]).reshape(1, -1)

        scores, recipe_ids = self._searchIndex(embedding)
        for score, recipe_id in zip(scores, recipe_ids):
            duplicate_id = cursor.duplicate(
                relation=trigger_context.relation,
                identifier=trigger_context.identifier,
            )
            cursor.set(
                relation=trigger_context.relation,
                identifier=duplicate_id,
                key_values={"recipe_id": recipe_id, "recipe_score": score},
            )



import logging
import os
import shutil
import typing
import uuid

import colorlog
import uvicorn

from motion.api import create_fastapi_app
from motion.client import ClientConnection
from motion.store import Store
from motion.utils import PRODUCTION_SESSION_ID


def create_token() -> str:
    """Creates a token for the API.

    Returns:
        str: The token.
    """
    return str(os.urandom(20).hex())


def create_example_app(name: str, author: str) -> None:
    """Creates a motion app from examples."""
    name = name.strip().lower()

    if name == "cooking":
        # Copy the example project
        shutil.copytree(
            os.path.join(os.path.dirname(__file__), f"examples/{name}"), name
        )

        # Create config
        with open(os.path.join(name, "mconfig.py"), "w") as f:
            f.write(
                open(
                    os.path.join(
                        os.path.dirname(__file__),
                        f"examples/{name}/mconfig.py",
                    ),
                )
                .read()
                .replace("{1}", author)
            )

    else:
        raise ValueError(f"Example application {name} does not exist.")


def create_app(name: str, author: str) -> None:
    """Creates a motion app."""
    name = name.strip().lower()
    if len(name.split(" ")) > 1:
        raise ValueError("Name cannot contain spaces.")

    if os.path.exists(name):
        raise ValueError(f"Directory {name} already exists.")

    # Copy over the example project
    shutil.copytree(os.path.join(os.path.dirname(__file__), "examples/basic"), name)

    # Create store setup file
    with open(os.path.join(name, "mconfig.py"), "w") as f:
        f.write(
            open(
                os.path.join(
                    os.path.dirname(__file__),
                    "examples/basic/mconfig.py",
                ),
            )
            .read()
            .replace("{0}", name)
            .replace("{1}", author)
        )


def init(
    mconfig: dict, disable_cron_triggers: bool = False, session_id: str = ""
) -> Store:
    """Initializes a motion application by creating a data store and adding relations and triggers.

    Args:
        mconfig (dict): The motion configuration.
        disable_cron_triggers (bool, optional): Whether to disable cron triggers. Used during testing. Defaults to False.
        prod (bool, optional): Whether to run in production mode. Defaults to False.

    Returns:
        Store: The motion store.
    """
    # TODO(shreyashankar): use version to check for updates when
    # needing to reinit

    try:
        name = mconfig["application"]["name"]
        author = mconfig["application"]["author"]
    except Exception as e:
        raise Exception("Motion config must have application name and author.")

    checkpoint = mconfig["checkpoint"] if "checkpoint" in mconfig else "0 * * * *"

    if session_id == "":
        session_id = str(uuid.uuid4())

    MOTION_HOME = os.environ.get("MOTION_HOME", os.path.expanduser("~/.cache/motion"))

    store = Store(
        name,
        session_id=session_id,
        datastore_prefix=os.path.join(MOTION_HOME, "datastores"),
        checkpoint=checkpoint,
        disable_cron_triggers=disable_cron_triggers,
    )

    # Create relations
    for relation in mconfig["relations"]:
        store.addrelation_pa(relation.__name__, relation)

    # Create triggers
    for trigger in mconfig["triggers"]:
        params = mconfig.get("trigger_params", {}).get(trigger, {})

        store.addTrigger(
            name=trigger.__name__,
            trigger=trigger,
            params=params,
        )

    # Start store
    store.start()

    return store


def serve(
    mconfig: dict,
    host: str = "0.0.0.0",
    port: int = 5000,
    motion_logging_level: str = "INFO",
) -> None:
    """Serves a Motion application. Uses the MOTION_API_TOKEN environment variable to authenticate API requests.

    Args:
        mconfig (dict): The motion config, found in the mconfig.py file.
        host (str, optional): The host to serve on. Defaults to "0.0.0.0".
        port (int, optional): The port to serve on. Defaults to 5000.
        motion_logging_level (str, optional): The logging level for motion. Can be one of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL". Defaults to "INFO".
    """
    configureLogging(motion_logging_level)
    store = init(mconfig, session_id=PRODUCTION_SESSION_ID)
    serve_store(store, host, port)


def serve_store(store: Store, host: str, port: int) -> None:
    # Start fastapi server
    app = create_fastapi_app(store)
    uvicorn.run(app, host=host, port=port)


def configureLogging(level: str) -> None:
    handler = logging.StreamHandler()

    formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(asctime)s %(levelname)-8s%(reset)s %(blue)s%(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "bold_red",
        },
    )
    handler.setFormatter(formatter)
    logger = logging.getLogger("motion")
    logger.addHandler(handler)
    logger.setLevel(level)


def test(
    mconfig: dict,
    wait_for_triggers: typing.List[str] = [],
    disable_cron_triggers: bool = False,
    motion_logging_level: str = "WARNING",
    session_id: str = "",
) -> ClientConnection:
    """Creates a test connection to a Motion application, defined by a mconfig. This will run the application and then shut it down when the connection goes out of scope. Uses the MOTION_API_TOKEN environment variable to authenticate.

    Args:
        mconfig (dict): Config for the Motion application, found in the mconfig.py file.
        wait_for_triggers (typing.List[str], optional): List of cron-scheduled trigger names to wait for a first completion of. Typically used to wait for a first scrape of data.
        disable_cron_triggers (bool, optional): Whether cron triggers should be disabled for this session (can speed up testing some non-cron triggers). Defaults to False. Cannot be True if wait_for_triggers is not empty.
        motion_logging_level (str, optional): Logging level for motion. Can be one of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL". Defaults to "WARNING". Use "INFO" if you want to see all trigger execution logs.
        session_id (str, optional): Session ID to use for this connection. Defaults to a random UUID if empty.

    Returns:
        connection (motion.ClientConnection): A cannection to the Motion application.
    """
    if wait_for_triggers and disable_cron_triggers:
        raise ValueError("Cannot wait for triggers if cron triggers are disabled.")

    configureLogging(motion_logging_level)
    store = init(
        mconfig,
        disable_cron_triggers=disable_cron_triggers,
        session_id=session_id,
    )
    app = create_fastapi_app(store, testing=True)

    connection = ClientConnection(
        mconfig["application"]["name"],
        server=app,
        bearer_token=os.environ.get("MOTION_API_TOKEN", ""),
    )
    connection.addStore(store)

    for trigger in wait_for_triggers:
        connection.waitForTrigger(trigger)

    return connection


def connect(
    name: str,
    host: str = "0.0.0.0",
    port: int = 5000,
    wait_for_triggers: typing.List[str] = [],
    motion_api_token: str = "",
) -> ClientConnection:
    """Connects to a Motion application that is already being served.

    Args:
        name (str): The name of the Motion application.
        host (str, optional): The host of the Motion application. Defaults to localhost.
        port (int, optional): The port of the Motion application. Defaults to 5000.
        wait_for_triggers (typing.List[str], optional): List of cron-scheduled trigger names to wait for a first completion of. Typically used to wait for a first scrape of data.
        motion_api_token (str, optional): API token set as the environment variable on the host serving the Motion application. If not provided as an argument, the token will be read from environment (possibly throwing an error if the environment doesn't have an API token defined).

    Returns:
        connection (motion.ClientConnection): A connection to the Motion application.
    """
    #  Check logs
    MOTION_API_TOKEN = (
        motion_api_token if motion_api_token else os.environ.get("MOTION_API_TOKEN", "")
    )

    server = "http://" + host + ":" + str(port)
    connection = ClientConnection(name, server, bearer_token=MOTION_API_TOKEN)
    for trigger in wait_for_triggers:
        connection.waitForTrigger(trigger)

    return connection



"""
Database connection, with functions that a users is allowed to
call within trigger lifecycle methods.
"""


import collections
import threading
import typing
import uuid
from enum import Enum

import duckdb
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from motion.utils import TriggerElement, TriggerFn, logger


class Cursor:
    """A connection to a Motion data store, only accessible within Motion triggers."""

    def __init__(
        self,
        *,
        name: str,
        relations: typing.Dict[str, pa.Table],
        log_table: pa.Table,
        table_columns: typing.Dict[str, list[str]],
        triggers: typing.Dict[str, list[TriggerFn]],
        write_lock: threading.Lock,
        session_id: str,
        wait_for_results: bool = False,
        triggers_to_run_on_duplicate: typing.Dict[TriggerFn, TriggerElement] = {},
        spawned_by: TriggerElement = None,  # type: ignore
    ):
        self.name = name
        self.relations: typing.Dict[str, pa.Table] = relations
        self.log_table = log_table
        self.table_columns = table_columns
        self.triggers = triggers
        self.write_lock = write_lock
        self.session_id = session_id
        self.wait_for_results = wait_for_results
        self.fit_events: list[threading.Event] = []
        self.triggers_to_run_on_duplicate = triggers_to_run_on_duplicate
        self.spawned_by = spawned_by

    def __del__(self) -> None:
        if self.wait_for_results:
            self.waitForResults()

    def waitForResults(self) -> None:
        """Waits for all fit events to finish."""
        for t in self.fit_events:
            t.wait()
        self.fit_events = []

    def getNewId(self, relation: str, key: str = "identifier") -> str:
        """Gets a new id for a relation.

        Args:
            relation (str): The relation to get the new id for.
            key (str, optional): The key to get the new id for. Defaults to "id".

        Returns:
            str: The new id.
        """
        new_id = str(uuid.uuid4())

        # Check if the id already exists
        if self.exists(relation, new_id):
            return self.getNewId(relation, key)

        return new_id

    def exists(
        self,
        relation: str,
        identifier: str,
    ) -> bool:
        """Determines if a record exists in a relation.

        Args:
            relation (str): The relation to check.
            identifier (str): The record's identifier.

        Returns:
            bool: True if the record exists, False otherwise.
        """
        if relation not in self.relations.keys():
            raise KeyError(f"Relation {relation} does not exist.")

        # Check if identifier exists in pyarrow table
        table = self.relations[relation]
        condition = pc.equal(table["identifier"], identifier)
        mask = pc.filter(table["identifier"], condition)
        result = len(mask) > 0

        return result

    def set(
        self,
        *,
        relation: str,
        key_values: typing.Dict[str, typing.Any],
        identifier: str = "",
    ) -> str:
        """Sets given key-value pairs for an identifier in a relation.
        Overwrites existing values.

        Args:
            relation (str): The relation to set the value in.
            identifier (str): The id of the record to set the value for.
            key_values (typing.Dict[str, typing.Any]): The key-value pairs to set.

        Returns:
            str: The identifier of the record.
        """
        if relation is None:
            raise ValueError("relation cannot be None.")

        if relation not in self.relations:
            raise KeyError(f"Relation {relation} does not exist.")

        exists = True
        if not identifier:
            identifier = self.getNewId(relation)
            exists = False

        if exists:
            exists = self.exists(relation, identifier)

        # Convert enums to their values
        for key, value in key_values.items():
            if isinstance(value, Enum):
                key_values.update({key: value.value})

        # Insert or update based on identifier
        with self.write_lock:
            table = self.relations[relation]

            if not exists:
                new_row_dict = {n: None for n in table.schema.names}
                new_row_dict.update(
                    {
                        "identifier": identifier,  # type: ignore
                        "create_at": pd.Timestamp.now(),  # type: ignore
                        "session_id": self.session_id,  # type: ignore
                    }
                )
                new_row_dict.update(key_values)
                new_row_df = pd.DataFrame(new_row_dict, index=[0])
                new_row = pa.Table.from_pandas(new_row_df, schema=table.schema)

                # Check schemas match
                if collections.Counter(new_row.schema.names) != collections.Counter(
                    new_row_df.columns.values
                ):
                    raise AttributeError(
                        f"One of the keys you are trying to set is not a valid key in the relation {relation}. Please double check your keys."
                    )

                final_table = pa.concat_tables([table, new_row])
                self.relations[relation] = final_table

            else:
                condition = pc.equal(table["identifier"], identifier)
                row = pc.filter(table, condition).to_pandas()

                for key, value in key_values.items():
                    row.at[0, key] = value

                new_row = pa.Table.from_pandas(row, schema=table.schema)

                filtered_table = pc.filter(table, pc.invert(condition))
                final_table = pa.concat_tables([filtered_table, new_row])
                self.relations[relation] = final_table

        # Get all triggers to run
        triggers_to_run: typing.Dict[TriggerFn, TriggerElement] = {}
        for key, value in key_values.items():
            if f"{relation}.{key}" not in self.triggers:
                continue

            trigger_context = TriggerElement(
                relation=relation,
                identifier=identifier,
                key=key,
                value=value,
            )
            for trigger in self.triggers.get(f"{relation}.{key}", []):
                if trigger in triggers_to_run.keys():
                    continue

                triggers_to_run[trigger] = trigger_context

        # Run triggers, passing in remainder of triggers_to_run
        for trigger, trigger_context in triggers_to_run.items():
            other_triggers_to_run = {
                k: triggers_to_run[k] for k in triggers_to_run if k != trigger
            }
            self.executeTrigger(
                trigger=trigger,
                trigger_context=trigger_context,
                triggers_to_run_on_duplicate=other_triggers_to_run,
            )

        return identifier

    def logTriggerExecution(
        self,
        trigger_name: str,
        trigger_version: int,
        trigger_action: str,
        trigger_context: TriggerElement,
    ) -> None:
        """Logs the execution of a trigger.

        Args:
            trigger_name (str): The name of the trigger.
            trigger_version (int): The version of the trigger.
            trigger_action (str): The action of the trigger.
            trigger_context (TriggerElement): The element that triggered the trigger.
        """

        # Append to the log table
        new_row = {
            "executed_time": pd.Timestamp.now(),
            "session_id": self.session_id,
            "trigger_name": trigger_name,
            "trigger_version": trigger_version,
            "trigger_action": trigger_action,
            "relation": trigger_context.relation,
            "identifier": trigger_context.identifier,
            "trigger_key": trigger_context.key,
        }
        new_row = pa.Table.from_pandas(
            pd.DataFrame(new_row, index=[0]), schema=self.log_table.schema
        )

        with self.write_lock:
            self.log_table = pa.concat_tables([self.log_table, new_row])

    def executeTrigger(
        self,
        *,
        trigger: TriggerFn,
        trigger_context: TriggerElement,
        triggers_to_run_on_duplicate: typing.Dict[TriggerFn, TriggerElement] = {},
    ) -> None:
        """Executes a trigger, logging completion of infer and fit methods.

        Args:
            trigger (TriggerFn): The trigger to execute.
            trigger_context (TriggerElement): The element that triggered the trigger.
            triggers_to_run (typing.Dict[TriggerFn, TriggerElement], optional): The triggers to run whenever duplicate is called within a trigger. Defaults to {}.
        """
        try:
            self._executeTrigger(trigger, trigger_context, triggers_to_run_on_duplicate)
        except RecursionError:
            raise RecursionError(
                f"Recursion error in trigger {trigger[0]}. Please make sure you do not have a cycle in your triggers."
            )

    def _executeTrigger(
        self,
        trigger: TriggerFn,
        trigger_context: TriggerElement,
        triggers_to_run_on_duplicate: typing.Dict[TriggerFn, TriggerElement] = {},
    ) -> None:
        """Execute a trigger.

        Args:
            trigger (TriggerFn): The trigger to execute.
            trigger_context (TriggerElement): The element that triggered the trigger.
            triggers_to_run (typing.Dict[TriggerFn, TriggerElement], optional): The triggers to run whenever duplicate is called within a trigger. Defaults to {}.
        """
        trigger_name, trigger_fn = trigger
        logger.info(
            f"Running trigger {trigger_name} for identifier {trigger_context.identifier}, key {trigger_context.key}..."
        )
        new_connection = Cursor(
            name=self.name,
            relations=self.relations,
            log_table=self.log_table,
            table_columns=self.table_columns,
            triggers=self.triggers,
            write_lock=self.write_lock,
            session_id=self.session_id,
            wait_for_results=self.wait_for_results,
            triggers_to_run_on_duplicate=triggers_to_run_on_duplicate,
            spawned_by=trigger_context,
        )

        # Get route for key
        route = trigger_fn.route_map.get(
            f"{trigger_context.relation}.{trigger_context.key}", None
        )
        if route is None:
            raise NotImplementedError(
                f"Route not found for {trigger_context.relation}.{trigger_context.key}."
            )

        # Execute the transform lifecycle: infer -> fit
        infer_context = None
        if route.infer is not None:
            infer_context = route.infer(new_connection, trigger_context)
            self.logTriggerExecution(
                trigger_name, trigger_fn.version, "infer", trigger_context
            )

        # Fit is asynchronous
        if route.fit is not None:
            fit_thread = trigger_fn.fitWrapper(
                new_connection, trigger_name, trigger_context, infer_context
            )
            self.fit_events.append(fit_thread)
        else:
            logger.info(
                f"Finished running trigger {trigger_name} for identifier {trigger_context.identifier}."
            )

    def duplicate(self, relation: str, identifier: str) -> str:
        """Duplicates a record in a relation. Doesn't rerun any triggers for old keys on the new record.

        Args:
            relation (str): The relation to duplicate the record in.
            identifier (str): The identifier of the record to duplicate.

        Returns:
            str: The new identifier of the duplicated record.
        """
        new_id = self.getNewId(relation)

        with self.write_lock:
            table = self.relations[relation]
            condition = pc.equal(table["identifier"], identifier)

            row = pc.filter(table, condition).to_pandas()
            row.at[0, "identifier"] = new_id
            row.at[0, "derived_id"] = identifier

            new_row = pa.Table.from_pandas(row, schema=table.schema)

            final_table = pa.concat_tables([table, new_row])
            self.relations[relation] = final_table

        # Run triggers on duplicate if it was from element that spawned
        # the trigger in the first place
        if (
            self.spawned_by is not None
            and self.spawned_by.identifier == identifier
            and self.spawned_by.relation == relation
        ):
            for (
                trigger,
                trigger_context,
            ) in self.triggers_to_run_on_duplicate.items():
                other_triggers_to_run = {
                    k: self.triggers_to_run_on_duplicate[k]
                    for k in self.triggers_to_run_on_duplicate
                    if k != trigger
                }
                new_trigger_context = trigger_context._replace(identifier=new_id)
                self.executeTrigger(
                    trigger=trigger,
                    trigger_context=new_trigger_context,
                    triggers_to_run_on_duplicate=other_triggers_to_run,
                )

        return new_id

    def get(
        self,
        *,
        relation: str,
        identifier: str,
        keys: list[str],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Get values for an identifier's keys in a relation. Can pass in ["*"] as the keys argument to get all keys.

        Args:
            relation (str): The relation to get the value from.
            identifier (str): The identifier of the record to get the value for.
            keys (typing.List[str]): The keys to get the values for.

        Keyword Args:
            include_derived (bool, optional): Whether to include derived ids in the result. Defaults to False.
            filter_null (bool, optional): Whether to filter out null values. Filters all records with any null walue for any of the keys requested. Only used in conjuction with include_derived. Defaults to True.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to False.

        Returns:
            typing.Any: The values for the keys.
        """

        if not identifier:
            raise ValueError("Identifier cannot be empty.")

        if keys == ["*"]:
            keys = self.relations[relation].schema.names

        if not keys or not all(
            [k in self.relations[relation].schema.names for k in keys]
        ):
            raise ValueError(f"Not all keys {keys} not found in relation {relation}.")

        con = duckdb.connect()
        scanner = pa.dataset.Scanner.from_dataset(
            pa.dataset.dataset(self.relations[relation])
        )

        if not kwargs.get("include_derived", False):
            res = con.execute(
                f"SELECT {', '.join(keys)} FROM scanner WHERE identifier = '{identifier}'"
            ).fetchall()

            if not res:
                return res

            res = res[0]
            res_dict = {k: v for k, v in zip(keys, res)}
            res_dict.update({"identifier": identifier})

            return (
                pd.Series(res_dict).to_frame().T
                if kwargs.get("as_df", False)
                else res_dict
            )

        # Recursively get derived ids
        id_res = con.execute(
            f"SELECT identifier FROM scanner WHERE derived_id = '{identifier}'"
        ).fetchall()
        id_res = [i[0] for i in id_res]
        all_ids = [identifier] + id_res
        while len(id_res) > 0:
            id_res_str = [f"'{str(i)}'" for i in id_res]
            id_res = con.execute(
                f"SELECT identifier FROM scanner WHERE derived_id IN ({', '.join(id_res_str)})"
            ).fetchall()
            id_res = [i[0] for i in id_res]
            all_ids.extend(id_res)

        if "identifier" not in keys:
            keys.append("identifier")

        kwargs.update({"include_derived": False})
        return self.mget(
            relation=relation,
            identifiers=all_ids,
            keys=keys,
            **kwargs,
        )

    def _get_derived_ids(
        self, con: duckdb.DuckDBPyConnection, identifier: str
    ) -> list[str]:
        """Get all derived ids for an identifier.

        Args:
            con (duckdb.PyConnection): The connection to use.
            identifier (str): The identifier to get the derived ids for.

        Returns:
            typing.List[str]: The derived ids.
        """
        id_res = con.execute(
            f"SELECT identifier FROM scanner WHERE derived_id = '{identifier}'"
        ).fetchall()
        id_res = [i[0] for i in id_res]
        all_ids = [identifier] + id_res
        while len(id_res) > 0:
            id_res_str = [f"'{str(i)}'" for i in id_res]
            id_res = con.execute(
                f"SELECT identifier FROM scanner WHERE derived_id IN ({', '.join(id_res_str)})"
            ).fetchall()
            id_res = [i[0] for i in id_res]
            all_ids.extend(id_res)
        return all_ids

    def mget(
        self,
        *,
        relation: str,
        identifiers: list[str],
        keys: list[str],
        **kwargs: typing.Any,
    ) -> pd.DataFrame:
        """Get values for a many identifiers' keys in a relation. Can pass in ["*"] as the keys argument to get all keys.

        Args:
            relation (str): The relation to get the value from.
            identifiers (typing.List[int]): The ids of the records to get the value for.
            keys (typing.List[str]): The keys to get the values for.

        Keyword Args:
            include_derived (bool, optional): Whether to include derived ids in the result. Defaults to False.
            filter_null (bool, optional): Whether to filter out null values. Filters all records with any null walue for any of the keys requested. Only used in conjuction with include_derived. Defaults to True.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to False.


        Returns:
            pd.DataFrame: The values for the key.
        """

        con = duckdb.connect()
        scanner = pa.dataset.Scanner.from_dataset(
            pa.dataset.dataset(self.relations[relation])
        )

        # Find all derived ids
        if kwargs.get("include_derived", False):
            all_derived_ids = []
            for identifier in identifiers:
                all_derived_ids.extend(self._get_derived_ids(con, identifier))
            identifiers = list(set(identifiers + all_derived_ids))

        all_ids_str = [f"'{str(i)}'" for i in identifiers]

        if keys == ["*"]:
            keys = self.relations[relation].schema.names

        if "identifier" not in keys:
            keys.append("identifier")
        res = con.execute(
            f"SELECT {', '.join(keys)} FROM scanner WHERE identifier IN ({', '.join(all_ids_str)})"
        ).fetch_arrow_table()

        if kwargs.get("filter_null", True):
            if "derived_id" not in keys:
                res = pa.compute.drop_null(res).combine_chunks()

            else:
                # Filter out rows where all columns except derived_id are null
                keep_null_idx = res.schema.get_field_index("derived_id")
                col_indices = [
                    i for i in range(len(res.schema.names)) if i != keep_null_idx
                ]
                valid_cols = [pa.compute.is_valid(res.column(i)) for i in col_indices]
                cond = valid_cols[0]
                for i in range(1, len(valid_cols)):
                    cond = pa.compute.and_(cond, valid_cols[i])
                res = res.filter(cond)

        res = res.to_pandas()
        if kwargs.get("as_df", False):
            return res

        return res.to_dict("records")

    def getIdsForKey(self, relation: str, key: str, value: typing.Any) -> list[int]:
        """Get ids for a key-value pair in a relation.

        Args:
            relation (str): The relation to get the value from.
            key (str): The key to get the values for.
            value (typing.Any): The value to get the ids for.

        Returns:
            typing.List[int]: The ids for the key-value pair.
        """
        con = duckdb.connect()
        scanner = pa.dataset.Scanner.from_dataset(
            pa.dataset.dataset(self.relations[relation])
        )
        # table = self.relations[relation]

        res = con.execute(
            f"SELECT identifier FROM scanner WHERE {key} = ?",
            (value,),
        ).fetchall()
        return [r[0] for r in res]

    def sql(self, query: str, as_df: bool = True) -> typing.Any:
        """Executes a SQL query on the relations. Specify the relations as tables in the query.

        Args:
            query (str): SQL query to execute.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to True.

        Returns:
            typing.Any: Pandas dataframe if as_df is True, else list of tuples.
        """
        con = duckdb.connect()

        # Create a table for each relation
        for relation, table in self.relations.items():
            locals()[relation] = table

        return (
            con.execute(query).fetch_arrow_table().to_pandas()
            if as_df
            else con.execute(query).fetchall()
        )



import inspect
from typing import Callable, Dict, Union

from pydantic import BaseModel, Field, root_validator

from motion.trigger import Trigger


class Route(BaseModel):
    """A route is a mapping from a relation and key to trigger infer and fit methods. See the trigger [life cycle](/concepts/trigger) for more information.

    Args:
        relation (str): The relation to which this route applies.
        key (str): The key to which this route applies.
        infer (Union[Callable, None], optional): The method to call on infer. Typically a method of the trigger (e.g., `self.infer`). If None, no infer function will be called. Defaults to None.
        fit (Union[Callable, None], optional): The method to call on fit. Typically a method of the trigger (e.g., `self.fit`). If None, no fit function will be called. Runs asynchronously. Defaults to None.
    """

    relation: str = Field(..., description="The relation to which this route applies.")
    key: str = Field(..., description="The key to which this route applies.")
    infer: Union[Callable, None] = Field(
        None,
        description="The method to call on infer. Typically a method of the trigger (e.g., `self.infer`). If None, no infer function will be called.",
    )
    fit: Union[Callable, None] = Field(
        None,
        description="The method to call on fit. Typically a method of the trigger (e.g., `self.fit`). If None, no fit function will be called. Runs asynchronously.",
    )

    @root_validator()
    def key_has_no_periods(cls, values: Dict) -> Dict:
        relation = values.get("relation")
        key = values.get("key")

        if "." in key:  # type: ignore
            raise ValueError(f"Route key {key} cannot contain periods.")
        if "." in relation:  # type: ignore
            raise ValueError(f"Route relation {relation} cannot contain periods.")

        return values

    def validateTrigger(self, trigger_object: Trigger) -> None:
        if self.infer is not None:
            if getattr(trigger_object, self.infer.__name__, None) is None:
                raise ValueError(
                    f"Trigger {trigger_object.name} does not have an infer function named {self.infer.__name__}."
                )

            if len(inspect.signature(self.infer).parameters) != 2:
                raise ValueError(
                    f"Infer method {self.infer.__name__} should have 2 arguments: cursor and trigger_context."
                )

        if self.fit is not None:
            if getattr(trigger_object, self.fit.__name__, None) is None:
                raise ValueError(
                    f"Trigger {trigger_object.name} does not have a fit function named {self.fit.__name__}."
                )

            if len(inspect.signature(self.fit).parameters) != 3:
                raise ValueError(
                    f"Fit method {self.fit.__name__} should have 3 arguments: cursor, trigger_context, and infer_context."
                )



import copy
import logging
from collections import namedtuple

logger = logging.getLogger(__name__)

TriggerElement = namedtuple(
    "TriggerElement", ["relation", "identifier", "key", "value"]
)
TriggerFn = namedtuple("TriggerFn", ["name", "fn"])

PRODUCTION_SESSION_ID = "PRODUCTION"


def update_params(mconfig: dict, params: dict) -> dict:
    """Updates the mconfig with the new trigger params.

    Args:
        mconfig (dict): config dict in mconfig.py
        params (dict): dict from trigger name to updated params.

    Returns:
        dict: new mconfig with updated params.
    """
    cp = copy.deepcopy(mconfig)
    cp["trigger_params"] = cp.get("trigger_params", {})

    for trigger_name, trigger_params in params.items():
        exists = False
        for trigger, existing_params in cp["trigger_params"].items():
            if trigger.__name__ == trigger_name:
                existing_params.update(trigger_params)
                exists = True
                break
        if not exists:
            cp["trigger_params"][trigger_name] = trigger_params

    return cp



import inspect
import typing
from collections import namedtuple
from datetime import date, datetime
from enum import Enum

import numpy as np
import pyarrow as pa
from pydantic import BaseModel, Extra

Field = namedtuple("Field", ["name", "type_"])


def type_to_name(t: type) -> str:
    return t.__name__.split(".")[-1].upper()


TYPE_TO_PA_TYPE = {
    int: pa.int64(),
    str: pa.string(),
    float: pa.float64(),
    bool: pa.bool_(),
    date: pa.date32(),
    datetime: pa.timestamp("us"),
    bytes: pa.binary(),
}


def get_arrow_type(t: type) -> pa.DataType:
    if t in TYPE_TO_PA_TYPE.keys():
        return TYPE_TO_PA_TYPE[t]

    elif inspect.isclass(t) and issubclass(t, Enum):
        # return pa_ext.EnumType.from_enum(t)
        return pa.string()

    elif typing.get_origin(t) == list:
        sub_t = typing.get_args(t)[0]
        return pa.list_(get_arrow_type(sub_t))

    elif typing.get_origin(t) == dict:
        kt = typing.get_args(t)[0]
        vt = typing.get_args(t)[1]
        return pa.map_(get_arrow_type(kt), get_arrow_type(vt))

    # Check if numpy type
    elif issubclass(t, np.generic) or issubclass(t, np.ndarray):
        return pa.from_numpy_dtype(t)

    # TODO: Add support for other types
    raise TypeError(f"Type {t} not supported.")


class Schema(BaseModel, extra=Extra.allow):
    """Schema for a Motion relation. Read more about relations in Motion [here](/concepts/relation).

    Example:

    ```python
    class User(Schema):
        name: str
        dob: date
        hometown: str
    ```

    All schemas also have the following fields, by default:
    - `identifier` (str): A unique identifier for each record in the relation.
    - `create_at` (datetime): The time at which the record was created.
    - `derived_id` (str): The identifier of the record that was derived from, if any.
    - `session_id` (str): The identifier of the session that created the record.

    Raises:
        TypeError: If a type is not supported. Supported types are: int, str, float, bool, date, datetime, bytes, list, dict, numpy types, and enums.
    """

    identifier: str
    derived_id: str
    create_at: datetime
    session_id: str

    @classmethod
    def formatPaSchema(cls, relation: str) -> pa.Schema:
        """Formats a pyarrow schema for the given relation. This is used internally by Motion.

        Returns:
            pa.Schema: Schema for table based on annotations.
        """

        fields = [Field(key, val) for key, val in cls.__annotations__.items()]
        user_defined_fields = [
            f
            for f in fields
            if not f.name == "identifier"
            and not f.name == "create_at"
            and not f.name == "derived_id"
            and not f.name == "session_id"
        ]
        pa_fields = [
            pa.field("identifier", pa.string(), nullable=False),
            pa.field(
                "create_at",
                pa.timestamp("us"),
                nullable=False,
            ),
            pa.field("derived_id", pa.string()),
            pa.field("session_id", pa.string(), nullable=False),
        ]

        for field in user_defined_fields:
            try:
                arrow_type = get_arrow_type(field.type_)
            except TypeError as e:
                raise TypeError(f"Error in {relation}.{field.name}: {e}")
            pa_fields.append(pa.field(field.name, arrow_type))

        schema = pa.schema(pa_fields)
        return schema



import threading
import time
import typing
from datetime import datetime

from croniter import croniter

from motion.cursor import Cursor
from motion.utils import PRODUCTION_SESSION_ID, TriggerElement, TriggerFn, logger


class CronThread(threading.Thread):
    """Thread that executes a task on cron schedule."""

    def __init__(
        self,
        cron_expression: str,
        cursor: Cursor,
        trigger_fn: TriggerFn,
        checkpoint_fn: typing.Callable,
        first_run_event: threading.Event,
        session_id: str,
    ) -> None:
        threading.Thread.__init__(self)
        self.daemon = True
        self.cron_expression = cron_expression
        self.cur = cursor
        self.trigger_fn = trigger_fn
        self.checkpoint_fn = checkpoint_fn
        self.running = True
        self.first_run = True
        self.first_run_event = first_run_event
        self.session_id = session_id

    def run(self) -> None:
        while self.running:
            next_time = croniter(self.cron_expression, datetime.now()).get_next(
                datetime
            )
            delay = (next_time - datetime.now()).total_seconds()

            # Wait until the scheduled time
            if not self.first_run:
                if delay > 0:
                    time.sleep(delay)
                else:
                    continue

            # Run trigger
            trigger_context = TriggerElement(
                relation="_cron",
                identifier="SCHEDULED",
                key=self.cron_expression,
                value=None,
            )

            try:
                self.cur.executeTrigger(
                    trigger=self.trigger_fn,
                    trigger_context=trigger_context,
                )

                self.cur.waitForResults()
                logger.info(
                    f"Finished waiting for background task {self.trigger_fn.name}."
                )
            except Exception as e:
                if self.session_id == PRODUCTION_SESSION_ID:
                    logger.error(
                        f"Error while running task {self.trigger_fn.name}: {e}"
                    )
                    continue
                else:
                    if self.first_run:
                        self.first_run_event.set()
                    raise e

            if self.first_run:
                self.first_run_event.set()
                self.first_run = False

            self.checkpoint_fn()
            logger.info(f"Checkpointed store from task {self.trigger_fn.name}.")

    def stop(self) -> None:
        logger.info(f"Stopping task thread for name {self.trigger_fn.name}")
        self.running = False


class CheckpointThread(threading.Thread):
    def __init__(
        self,
        store_name: str,
        checkpoint_fn: typing.Callable,
        cron_expression: str,
    ) -> None:
        threading.Thread.__init__(self)
        self.name = store_name
        self.daemon = True
        self.checkpoint_fn = checkpoint_fn
        self.running = True
        self.cron_expression = cron_expression

    def run(self) -> None:
        while self.running:
            next_time = croniter(self.cron_expression, datetime.now()).get_next(
                datetime
            )
            delay = (next_time - datetime.now()).total_seconds()
            if delay > 0:
                time.sleep(delay)
            else:
                continue

            logger.info(f"Checkpointing store {self.name}")
            self.checkpoint_fn()
            logger.info(f"Finished checkpointing store {self.name}")

    def stop(self) -> None:
        logger.info(f"Stopping checkpoint thread for store {self.name}")
        self.running = False



from schemas import Query, Recipe
from triggers import ScrapeBonAppetit, SearchRecipe

MCONFIG = {
    "application": {
        "name": "cooking",
        "author": "{1}",
        "version": "0.1",
    },
    "relations": [Recipe, Query],
    "triggers": [SearchRecipe, ScrapeBonAppetit],
    "trigger_params": {SearchRecipe: {"num_recipe_results": 10}},
    "checkpoint": "0 * * * *",
}



import asyncio
import json
from typing import Any, Dict, List

import aiohttp
import pandas as pd
import requests
from bs4 import BeautifulSoup
from schemas import RecipeSource

import motion


class ScrapeBonAppetit(motion.Trigger):
    def routes(self) -> List[motion.Route]:
        return [
            motion.Route(
                relation="",
                key="0 * * * *",
                infer=self.scrape,
                fit=None,
            )
        ]

    def setUp(self, cursor: motion.Cursor) -> Dict:
        return {}

    def scrape(
        self, cursor: motion.Cursor, trigger_context: motion.TriggerElement
    ) -> None:
        # Set lower time bound to avoid scraping recipes we've already scraped
        lower_bound = cursor.sql(
            "SELECT MAX(create_at) AS lower_bound FROM Recipe WHERE src='Bon Appetit'"
        )["lower_bound"].values[0]
        if pd.isnull(lower_bound):
            lower_bound = "2019-01-01"
        else:
            lower_bound = str(lower_bound)

        url_list = asyncio.run(
            fetch_all_urls("https://www.bonappetit.com/sitemap.xml", lower_bound)
        )

        print(f"Found {len(url_list)} recipes.")

        recipes = asyncio.run(scrape_all_recipes(url_list))
        print(f"Scraped {len(recipes)} recipes.")

        for recipe in recipes:
            recipe.update({"src": RecipeSource.BONAPPETIT})
            cursor.set(relation="Recipe", identifier="", key_values=recipe)


async def async_scrape_recipe(url: str, session: aiohttp.ClientSession) -> Any:
    try:
        async with session.get(url) as response:
            if response.status != 200:
                print(
                    f"Failed to fetch the webpage {url}. Status code: {response.status}"
                )
                return None

            content = await response.content.read()
            soup = BeautifulSoup(content, "html.parser")

            # Extract json ld data
            data = json.loads(soup.find("script", type="application/ld+json").string)
            title = data["name"]
            description = data["description"]
            ingredients = data["recipeIngredient"]
            instructions = [i["text"] for i in data["recipeInstructions"]]
            image_url = data["image"][0]
            if not image_url.startswith("http"):
                raise Exception("Image url is not valid.")

            recipe = {
                "title": title,
                "description": description,
                "ingredients": ";".join(ingredients),
                "instructions": ";".join(instructions),
                "image_url": image_url,
                "recipe_url": url,
            }

            return recipe
    except Exception as e:
        print(f"Failed to scrape recipe: {url} with error {e}")


async def scrape_all_recipes(url_list: List[str]) -> Any:
    async with aiohttp.ClientSession() as session:
        tasks = [async_scrape_recipe(url, session) for url in url_list]

        ret = await asyncio.gather(*tasks)
        ret = [x for x in ret if x is not None]

    return ret


async def fetch_sitemap_urls(sitemap_url: str, session: aiohttp.ClientSession) -> Any:
    try:
        async with session.get(sitemap_url) as response:
            if response.status != 200:
                print(
                    f"Failed to fetch the sitemap {sitemap_url}. Status code: {response.status}"
                )
                return []

            content = await response.content.read()

            soup = BeautifulSoup(content, "xml")
            urls = [loc.text for loc in soup.find_all("loc")]

        return urls
    except Exception as e:
        print(f"Failed to fetch sitemap: {sitemap_url} with error {e}")


async def fetch_all_urls(sitemap_url: str, date_threshold: str) -> Any:
    content = requests.get(sitemap_url).content
    soup = BeautifulSoup(content, "xml")

    filtered_urls = []
    for sitemap in soup.find_all("sitemap"):
        lastmod_element = sitemap.find("lastmod")
        lastmod_date = lastmod_element.text

        if lastmod_date > date_threshold:
            loc_element = sitemap.find("loc")
            filtered_urls.append(loc_element.text)

    async with aiohttp.ClientSession() as session:
        # urls = await fetch_sitemap_urls(sitemap_url, session)
        tasks = [fetch_sitemap_urls(url, session) for url in filtered_urls]

        ret = await asyncio.gather(*tasks)
        ret = [x for x in ret if x is not None]
        ret = [item for sublist in ret for item in sublist]
        ret = [x for x in ret if "/recipe/" in x]

    return ret



MCONFIG = {
    "application": {
        "name": "{0}",
        "author": "{1}",
        "version": "0.1",
    },
    "relations": [],
    "triggers": [],
    "trigger_params": {},
    "checkpoint": "0 * * * *",
}



from pydantic import BaseModel, Extra


class GetRequest(BaseModel, extra=Extra.allow):
    relation: str
    identifier: str
    keys: list


class MgetRequest(BaseModel, extra=Extra.allow):
    relation: str
    identifiers: list
    keys: list


class PartialSetRequest(BaseModel):
    relation: str
    identifier: str = ""


class SetRequest(BaseModel):
    relation: str
    identifier: str = ""
    key_values: dict


class SqlRequest(BaseModel):
    query: str
    as_df: bool = True


class DuplicateRequest(BaseModel):
    relation: str
    identifier: str


class WaitRequest(BaseModel):
    trigger: str



from enum import Enum

import motion


class RecipeSource(Enum):
    BONAPPETIT = "Bon Appetit"


class Recipe(motion.Schema):
    src: RecipeSource
    title: str
    description: str
    ingredients: str
    instructions: str
    image_url: str
    recipe_url: str


class Query(motion.Schema):
    username: str
    ingredients: str
    recipe_id: str
    recipe_score: float
    feedback: bool



import io
import json
import typing
from enum import Enum

import pandas as pd
import requests
from fastapi import FastAPI, HTTPException
from fastapi.testclient import TestClient

from motion.store import Store


class ClientConnection:
    """A client connection to a Motion application."""

    def __init__(
        self,
        name: str,
        server: typing.Union[str, FastAPI],
        bearer_token: str,
    ) -> None:
        if not bearer_token:
            raise ConnectionError(
                f"Could not find bearer token for {name}. Please set the MOTION_API_TOKEN environment variable, or pass in the token to the `motion.connect` function if you are using `motion.connect`."
            )

        self.name = name

        self.request_headers = {"Authorization": f"Bearer {bearer_token}"}
        if isinstance(server, FastAPI):
            self.server = server

        else:
            self.server = "http://" + server  # type: ignore
            try:
                response = requests.get(self.server + "/ping/")  # type: ignore
                if response.status_code != 200:
                    raise Exception(
                        f"Could not successfully connect to server for {self.name}; getting status code {response.status_code}."
                    )
            except requests.exceptions.ConnectionError:
                raise Exception(
                    f"Could not connect to server for {self.name} at {self.server}. Please run `motion serve` first."
                )
            self.session_id = requests.get(self.server + "/session_id/").json()  # type: ignore

    def addStore(self, store: Store) -> None:
        self.store = store
        self.session_id = self.store.session_id

    def close(self, wait: bool = False) -> None:
        if isinstance(self.server, FastAPI):
            self.store.stop(wait=wait)

    def __del__(self) -> None:
        self.close(wait=False)

    def getWrapper(self, dest: str, **kwargs: typing.Any) -> typing.Any:
        if isinstance(self.server, FastAPI):
            with TestClient(self.server) as client:
                response = client.request(
                    "get", dest, json=kwargs, headers=self.request_headers
                )
        else:
            response = requests.get(
                self.server + dest, json=kwargs, headers=self.request_headers
            )

        if response.status_code != 200:
            raise HTTPException(
                status_code=response.status_code, detail=response.content
            )

        with io.BytesIO(response.content) as data:
            if response.headers["content-type"] == "application/octet-stream":
                df = pd.read_parquet(data, engine="pyarrow")
                return df

            if response.headers["content-type"] == "application/json":
                return json.loads(response.content)

    def postWrapper(
        self, dest: str, data: typing.Any, files: typing.Any = None
    ) -> typing.Any:
        if isinstance(self.server, FastAPI):
            with TestClient(self.server) as client:
                response = client.request(
                    "post",
                    dest,
                    data=data,
                    files=files,
                    headers=self.request_headers,
                )
        else:
            response = requests.post(
                self.server + dest,
                data=data,
                files=files,
                headers=self.request_headers,
            )

        if response.status_code != 200:
            raise HTTPException(
                status_code=response.status_code, detail=response.content
            )

        return response.json()

    def waitForTrigger(self, trigger: str) -> typing.Any:
        """Waits for a cron-scheduled trigger to complete its first run.

        Args:
            trigger (str): The name of the trigger.
        """
        return self.postWrapper("/wait_for_trigger/", data={"trigger": trigger})

    def get(
        self,
        *,
        relation: str,
        identifier: str,
        keys: list[str],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Get values for an identifier's keys in a relation. Can pass in ["*"] as the keys argument to get all keys. Wrapper for the cursor's get method.

        Args:
            relation (str): The relation to get the value from.
            identifier (str): The identifier of the record to get the value for.
            keys (typing.List[str]): The keys to get the values for.

        Keyword Args:
            include_derived (bool, optional): Whether to include derived ids in the result. Defaults to False.
            filter_null (bool, optional): Whether to filter out null values. Filters all records with any null walue for any of the keys requested. Only used in conjuction with include_derived. Defaults to True.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to False.

        Returns:
            typing.Any: The values for the keys.
        """
        args = {
            "relation": relation,
            "identifier": identifier,
            "keys": keys,
        }
        args.update(kwargs)

        response = self.getWrapper("/get/", **args)
        if not kwargs.get("as_df", False):
            response = response.to_dict(orient="records")
            if len(response) == 1:
                response = response[0]

        return response

    def mget(
        self,
        *,
        relation: str,
        identifiers: list[str],
        keys: list[str],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Get values for a many identifiers' keys in a relation. Can pass in ["*"] as the keys argument to get all keys. Wrapper for the cursor's mget method.

        Args:
            relation (str): The relation to get the value from.
            identifiers (typing.List[int]): The ids of the records to get the value for.
            keys (typing.List[str]): The keys to get the values for.

        Keyword Args:
            include_derived (bool, optional): Whether to include derived ids in the result. Defaults to False.
            filter_null (bool, optional): Whether to filter out null values. Filters all records with any null walue for any of the keys requested. Only used in conjuction with include_derived. Defaults to True.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to False.


        Returns:
            pd.DataFrame: The values for the key.
        """
        args = {
            "relation": relation,
            "identifiers": identifiers,
            "keys": keys,
        }
        args.update(kwargs)
        response = self.getWrapper("/mget/", **args)
        if not kwargs.get("as_df", False):
            return response.to_dict(orient="records")
        return response

    def set(
        self,
        *,
        relation: str,
        identifier: str,
        key_values: typing.Dict[str, typing.Any],
    ) -> typing.Any:
        """Sets given key-value pairs for an identifier in a relation.
        Overwrites existing values. Wrapper for the cursor's set method.

        Args:
            relation (str): The relation to set the value in.
            identifier (str): The id of the record to set the value for.
            key_values (typing.Dict[str, typing.Any]): The key-value pairs to set.

        Returns:
            str: The identifier of the record.
        """
        # Convert enums to their values

        for key, value in key_values.items():
            if isinstance(value, Enum):
                key_values.update({key: value.value})

        # Turn key-values into a dataframe to convert to parquet
        if identifier is None:
            identifier = ""

        df = pd.DataFrame(key_values, index=[0])
        memory_buffer = io.BytesIO()
        df.to_parquet(memory_buffer, engine="pyarrow", index=False)
        memory_buffer.seek(0)

        # Create request args
        args = {"args": json.dumps({"relation": relation, "identifier": identifier})}

        return self.postWrapper(
            "/set_python/",
            data=args,
            files={
                "key_values": (
                    "key_values",
                    memory_buffer,
                    "application/octet-stream",
                )
            },
        )

    def sql(self, *, query: str, as_df: bool = True) -> typing.Any:
        """Executes a SQL query on the relations. Specify the relations as tables in the query. Wrapper for the cursor's sql method.

        Args:
            query (str): SQL query to execute.
            as_df (bool, optional): Whether to return the result as a pandas dataframe. Defaults to True.

        Returns:
            typing.Any: Pandas dataframe if as_df is True, else list of tuples.
        """
        args = {"query": query, "as_df": as_df}

        return self.getWrapper("/sql/", **args)

    def duplicate(self, *, relation: str, identifier: str) -> typing.Any:
        """Duplicates a record in a relation. Doesn't rerun any triggers for old keys on the new record. Wrapper for the cursor's duplicate method.

        Args:
            relation (str): The relation to duplicate the record in.
            identifier (str): The identifier of the record to duplicate.

        Returns:
            str: The new identifier of the duplicated record.
        """
        data = json.dumps({"relation": relation, "identifier": identifier})
        return self.postWrapper(
            "/duplicate/",
            data=data,
        )

    def checkpoint(self) -> None:
        """Checkpoints the data store to disk. Wrapper for the store's checkpoint method. Takes no arguments and returns nothing."""
        self.postWrapper("/checkpoint/", data={})
